/*
 *  warning: this is just a sample server to permit testing the clients;
 * it is not as optimized or robust as it should be
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <string.h>
#include <time.h>

#include "errlib.h"
#include "sockwrap.h"

#define SRVPORT 1500

#define LISTENQ 15
#define MAXBUFL 255

#define MSG_ERR "wrong operands\r\n"
#define MSG_OVF "overflow\r\n"

#define MAX_UINT8T  0xff
#define MAX_UINT16T 0xffff
#define STATE_OK 0x00
#define STATE_V  0x01

#ifdef TRACE
#define trace(x) x
#else
#define trace(x)
#endif

char *prog;


int prot_a (int connfd)
{
  char buf[MAXBUFL+1]; /* +1 to make room for \0 */
  uint16_t op1, op2;
  uint32_t res;
  int nread;

  while (1)
  {
    trace( err_msg("(%s) - waiting for operands ...",prog) );
    if ((nread = Readline (connfd, buf, MAXBUFL)) == 0 )
      return 0;

    /* append the string terminator after CR-LF */

    buf[nread]='\0';
    trace( err_msg("(%s) --- received string '%s'",prog, buf) );

    /* get the operands */

    if (sscanf(buf,"%hu %hu",&op1,&op2) != 2) {
      trace( err_msg("(%s) --- wrong or missing operands",prog) );
      Write (connfd, MSG_ERR, strlen(MSG_ERR));
      continue;
    }
    trace( err_msg("(%s) --- operands %hu %hu",prog,op1,op2) );

    /* do the operation */

    res = op1 + op2;

    /* check for overflow */

    if (res > MAX_UINT16T) {
      trace( err_msg("(%s) --- sum overflow",prog) );
      Write (connfd, MSG_OVF, strlen(MSG_OVF));
      continue;
    }

    trace( err_msg("(%s) --- result of the sum: %lu", prog, res) );

    /* convert the result to a string */

    snprintf (buf, MAXBUFL, "%lu\r\n", res);

    /* send the result */

    Write (connfd, buf, strlen(buf) );
    trace( err_msg("(%s) --- result just sent back", prog) );
  }
}


int prot_b (int connfd)
{
  char buf[MAXBUFL];
  uint16_t op1, op2, res16;
  uint32_t res;
  int nread;

  while (1)
  {
    trace( err_msg("(%s) - waiting for operands ...",prog) );
    if ((nread = Readn (connfd, buf, 2*sizeof(op1))) != (2*sizeof(op1) ))
      return 0;

    /* get the operands */

    op1 = ntohs(* (uint16_t*)buf);
    op2 = ntohs(* (uint16_t*)(buf+2));

    /* do the operation */

    res = op1 + op2;

    /* check for overflow */

    if (res > MAX_UINT16T) {
      trace( err_msg("(%s) --- sum overflow",prog) );
      buf[0] = STATE_V;
      buf[1] = buf[2] = 0x00;
      Writen (connfd, buf, 3);
      continue;
    }

    trace( err_msg("(%s) --- result of the sum: %lu", prog, res) );

    /* convert the result to network order */

    res16 = htons((uint16_t)res);

    /* send the result */

    buf[0] = STATE_OK;
    Writen (connfd, buf, 1);
    Writen (connfd, (void*) &res16, sizeof(res16));
    trace( err_msg("(%s) --- result just sent back", prog) );
  }
}


int prot_t (int connfd)
{
  unsigned char buf[MAXBUFL]; /* unsigned IMPORTANT to avoid sign extension */
  uint16_t op1, op2, res16;
  uint32_t res;
  int nread, opsize;

  while (1)
  {
    trace( err_msg("(%s) - waiting for operands ...",prog) );
    if ((nread = Readn (connfd, buf, 1)) == 0 )
      return 0;

    /* get the length */

    if ((opsize=buf[0])!=1 && opsize!=2) {
      trace( err_msg("(%s) --- error: size %d not supported", prog, opsize) );
      return -1;
    }
    trace( err_msg("(%s) --- operands' size: %u bytes", prog, opsize) );

    /* get the operands */

    if ((nread = Readn (connfd, buf, 2*opsize)) != 2*opsize) {
      trace( err_msg( (nread==0) ?  "(%s) --- unexepected EOF - channel close by client?"
        : "(%s) --- 'Readn' interrupted: read less characters than expected",prog) );
      return -1;
    }

    if (opsize == 1) {
      op1 = buf[0];
      op2 = buf[1];
    }
    else {
      op1 = ntohs(* (uint16_t*)buf);
      op2 = ntohs(* (uint16_t*)(buf+2));
    }

    /* do the operation */

    res = op1 + op2;

    /* check for overflow */

    if (res > ((opsize==1) ? MAX_UINT8T : MAX_UINT16T)) {
      trace( err_msg("(%s) --- sum overflow",prog) );
      buf[0] = STATE_V;
      buf[1] = buf[2] = 0x00;
      Writen (connfd, buf, 1+opsize);
      continue;
    }

    /* send the result */

    trace( err_msg("(%s) --- result of the sum: %lu", prog, res) );

    buf[0] = STATE_OK;
    buf[1] = opsize;
    if (opsize == 1)
      buf[2] = (unsigned char)res;
    else {
      res16 = htons((uint16_t)res);
      memcpy (buf+2, &res16, 2);
    }
    Writen (connfd, buf, 2+opsize);
    trace( err_msg("(%s) --- result just sent back", prog) );
  }
}


int main (int argc, char *argv[])
{

  int listenfd, connfd, err=0;
  short port;
  struct sockaddr_in servaddr, cliaddr;
  socklen_t cliaddrlen = sizeof(cliaddr);

  /* for errlib to know the program name */

  prog = argv[0];

  /* check arguments */

  port=SRVPORT;

  if (argc!=2 || strlen(argv[1])!=2 || argv[1][0]!='-' ||
    (argv[1][1]!='a' && argv[1][1]!='b' && argv[1][1]!='t'))
    err_quit ("usage: %s <protocol>\n where <protocol> can be -a -b -t", prog);

  /* create socket */

  listenfd = Socket(AF_INET, SOCK_STREAM, 0);

  /* specify address to bind to */

  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(port);
  servaddr.sin_addr.s_addr = htonl (INADDR_ANY);

  Bind(listenfd, (SA*) &servaddr, sizeof(servaddr));

  trace (
    err_msg("(%s) socket created",prog);
    err_msg("(%s) listening on %s:%u", prog, inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port)) );

  Listen(listenfd, LISTENQ);

  while (1)
  {
    trace( err_msg ("(%s) waiting for connections ...", prog) );

    connfd = Accept (listenfd, (SA*) &cliaddr, &cliaddrlen);
    trace ( err_msg("(%s) - new connection from client %s:%u", prog, inet_ntoa(cliaddr.sin_addr), ntohs(cliaddr.sin_port)) );

    switch (argv[1][1])
    {
    case 'a':
      err = prot_a(connfd);
      break;
    case 'b':
      err = prot_b(connfd);
      break;
    case 't':
      err = prot_t(connfd);
      break;
    default:
      err_quit ("(%s) bug - unexpected case in a switch", prog);
    }

    Close (connfd);
    trace( err_msg ("(%s) - connection closed by %s", prog, (err==0)?"client":"server") );
  }
  return 0;
}
